/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright 2018, Olivier MATZ <zer0@droids-corp.org>
 */

/**
 * @defgroup ecoli_node_none None node
 * @ingroup ecoli_nodes
 * @{
 *
 * @brief A node that never matches anything.
 */

#pragma once

/** @} */
