/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright 2016, Olivier MATZ <zer0@droids-corp.org>
 */

/**
 * @defgroup ecoli_node_space Space node
 * @ingroup ecoli_nodes
 * @{
 *
 * @brief A node that matches whitespace.
 *
 * This node matches one string in the vector if it is only composed of
 * spaces, as interpreted by isspace().
 */

#pragma once

/* no API for now, since there is no specific configuration for this node */

/** @} */
